DAT
{
  GLOBALS
  Provided to help assist in SPIN translation

  32-bit-specific abs fn coerces argument to 32-bit signed value first
  #define abs32(x) labs((int32_t)(x))

  //Position tracking
  TARGET BSS _StartPointMantissa; //Brief target position of last move in queue
  TARGET BSS StartPointSteps; //Brief target position of last move in queue, expressed in steps
  TARGET BSS current_position; //Brief actual position of extruder head
  MOVE_STATE BSS move_state; //Brief numbers for tracking the current state of movement
}

PRI InitDDA

{
  Inititalise DDA movement structures
}

  'Set up default feedrate
  IF(F_StartPointMantissa == 0)

    'Per Bruce: The following assignment was the only reference to next_target.target.F, so it has been
    'commented out and removed from the actual assignment.  SEARCH_FEEDRATE_Z is defined in config.default.h          
    'F_StartPointMantissa = next_target.target.F = SEARCH_FEEDRATE_Z
    F_StartPointMantissa := SEARCH_FEEDRATE_Z

PRI NewStartPointDDA

{
  Distribute a new startpoint to DDA's internal structures without any movement.
  This is needed for example after homing or a G92. The new location must be in startpoint already.
}

  X_StartPointSteps := uM_DistanceToStepConversion(X_StartPointMantissa)
  Y_StartPointSteps := uM_DistanceToStepConversion(Y_StartPointMantissa)
  Z_StartPointSteps := uM_DistanceToStepConversion(Z_StartPointMantissa)
  E_StartPointSteps := uM_DistanceToStepConversion(E_StartPointMantissa)

PRI CreateDDA(DDA *dda, TARGET *target)
{
  CREATE a dda given current_position and a target, save to passed location so we can write directly into the queue
      \param *dda pointer to a dda_queue entry to overwrite
      \param *target the target position of this move

      \ref startpoint the beginning position of this move

      This function does a /lot/ of math. It works out directions for each axis, distance travelled, the time between the first and second step

      It also pre-fills any data that the selected accleration algorithm needs, and can be pre-computed for the whole move.

      This algorithm is probably the main limiting factor to print speed in terms of firmware limitations
 
  Regarding lookahead, we can distinguish everything into these cases:

  1.  Standard movement. To be joined with the previous move.
  2.  Movement after a pause. This interrupts lookahead, and invalidates
      prev_dda and prev_distance.
  3.  Non-move, e.g. a wait for temp. This also interrupts lookahead and makes
      prev_dda and prev_distance invalid. There might be more such cases in the
      future, e.g. when heater or fan changes are queued up, too.
  4.  Nullmove due to no movement expected, e.g. a pure speed change. This
      shouldn't interrupt lookahead and be handled af if the change would come
      with the next movement.
  5.  Nullmove due to movement smaller than a single step. Shouldn't interrupt
      lookahead either, but this small distance should be added to the next
      movement.
  6.  Lookahead calculation too slow. This is handled in dda_join_moves()
      already.  
}

  uint32_t steps
  uint32_t X_MantissaDifference
  uint32_t Y_MantissaDifference
  uint32_t Z_MantissaDifference
  uint32_t E_MantissaDifference

  uint32_t Distance
  uint32_t ClockLimit
  uint32_t ClockLimitCalculation

  'TODO: We should really make up a loop for all axes.
  'Think of what happens when a sixth axis (multi colour extruder) appears?
  X_MantissaDifference := (uint32_t)ABS32(X_EndPointMantissa - X_StartPointMantissa)
  Y_MantissaDifference := (uint32_t)ABS32(Y_EndPointMantissa - Y_StartPointMantissa)
  Z_MantissaDifference := (uint32_t)ABS32(Z_EndPointMantissa - Z_StartPointMantissa)

  X_Steps := uM_DistanceToStepConversion(X_EndPointMantissa)
  X_StepsDifference := ABS32(X_Steps - X_StartPointSteps)
  X_StartPointSteps := X_Steps

  Y_Steps := uM_DistanceToStepConversion(Y_EndPointMantissa)
  Y_StepsDifference := ABS32(Y_Steps - Y_StartPointSteps)
  Y_StartPointSteps := Y_Steps

  Z_Steps := uM_DistanceToStepConversion(Z_EndPointMantissa)
  Z_StepsDifference := ABS32(Z_Steps - Z_StartPointSteps)
  Z_StartPointSteps := Z_Steps

  X_Direction := (X_EndPointMantissa >= X_StartPointMantissa)?1:0
  Y_Direction := (Y_EndPointMantissa >= Y_StartPointMantissa)?1:0
  Z_Direction := (Z_EndPointMantissa >= Z_StartPointMantissa)?1:0

  IF(E_EndPointMantissa_relative)
        
    E_MantissaDifference := ABS32(E_EndPointMantissa)
    E_StepsDifference := ABS32(uM_DistanceToStepConversion(E_EndPointMantissa))
    E_Direction := (E_EndPointMantissa >= 0)?1:0
        
  ELSE
        
    E_MantissaDifference := (uint32_t)ABS32(E_EndPointMantissa - E_StartPointMantissa)
    E_Steps := uM_DistanceToStepConversion(E_EndPointMantissa)
    E_StepsDifference := ABS32(E_Steps - E_StartPointSteps)
    E_StartPointSteps := E_Steps
    E_Direction := (E_EndPointMantissa >= E_StartPointMantissa)?1:0         

  TotalSteps := X_StepsDifference
  FastAxisMantissaDifference := X_MantissaDifference
  FastAxisStepsPerMeter := X_STEPS_PER_METER

  IF(Y_StepsDifference > TotalSteps)
        
    TotalSteps := Y_StepsDifference
    FastAxisMantissaDifference := Y_MantissaDifference
    FastAxisStepsPerMeter := Y_STEPS_PER_METER        

  IF(Z_StepsDifference > TotalSteps)
        
    TotalSteps := Z_StepsDifference
    FastAxisMantissaDifference := Z_MantissaDifference
    FastAxisStepsPerMeter := Z_STEPS_PER_METER        

  IF(E_StepsDifference > TotalSteps)
        
    TotalSteps := E_StepsDifference
    FastAxisMantissaDifference := E_MantissaDifference
    FastAxisStepsPerMeter := E_STEPS_PER_METER         

  IF(TotalSteps == 0)
        
    NullMove := 1
        
  ELSE    
        
    'Get steppers ready to go
    PowerOn()
    StepperEnable()
    X_Enable()
    Y_Enable()
    'Z is enabled in StartDDA()
    E_Enable()

    'Since it's unusual to combine X, Y and Z changes in a single move on reprap,
    'check IF we can use simpler approximations before trying the full 3d approximation.
    IF(Z_MantissaDifference == 0)
                
      Distance := ApproximateDistance(X_MantissaDifference, Y_MantissaDifference)
                
    ELSEIF(X_MantissaDifference == 0 && Y_MantissaDifference == 0)
                
      Distance := Z_MantissaDifference
                
    ELSE
                
      Distance := ApproximateDistance3(X_MantissaDifference, Y_MantissaDifference, Z_MantissaDifference)                

    IF(Distance < 2)
                
      Distance := E_MantissaDifference                

    'Pre-calculate move speed in millimeter microseconds per step minute for less math in interrupt context
    'mm (Distance) * 60000000 us/min / step (total_steps) := mm.us per step.min
    'note: um (Distance) * 60000 == mm * 60000000
    'so in the interrupt we must simply calculate
    'mm.us per step.min / mm per min (F) := us per step

    'break this calculation up a bit and lose some precision because 300,000um * 60000 is too big for a uint32
    'calculate this with a uint64 IF you need the precision, but it'll take longer so routines with lots of short moves may suffer
    '2^32/6000 is about 715mm which should be plenty

    'changed * 10 to * (F_CPU / 100000) so we can work in cpu_ticks rather than microseconds.
    'timer.c SetTimer() routine altered for same reason

    'changed Distance * 6000 .. * F_CPU / 100000 to
    'Distance * 2400 .. * F_CPU / 40000 so we can move a Distance of up to 1800mm without overflowing
    uint32_t MoveDuration := ((Distance * 2400) / TotalSteps) * (F_CPU / 40000)

    'Similarly, find out how fast we can run our axes.
    'Do this for each axis individually, as the combined speed of two
    'or more axes can be higher than the capabilities of a single one.
                
    'TODO: Instead of calculating c_min directly, it's probably more simple
    'to calculate (maximum) MoveDuration for each axis, like done for
    'ACCELERATION_TEMPORAL above. This should make re-calculating the allowed F easier.

    ClockLimit := 0

    'check X axis
    ClockLimitCalculation := ((X_MantissaDifference * 2400L) / TotalSteps * (F_CPU / 40000) / MAXIMUM_FEEDRATE_X) << 8

    IF(ClockLimitCalculation > ClockLimit)
                
      ClockLimit := ClockLimitCalculation                
                
    'check Y axis
    ClockLimitCalculation := ((Y_MantissaDifference * 2400L) / TotalSteps * (F_CPU / 40000) / MAXIMUM_FEEDRATE_Y) << 8

    IF(ClockLimitCalculation > ClockLimit)
                
      ClockLimit := ClockLimitCalculation                
                
    'check Z axis
    ClockLimitCalculation := ((Z_MantissaDifference * 2400L) / TotalSteps * (F_CPU / 40000) / MAXIMUM_FEEDRATE_Z) << 8

    IF(ClockLimitCalculation > ClockLimit)
                
      ClockLimit := ClockLimitCalculation                
                
    'check E axis
    ClockLimitCalculation := ((E_MantissaDifference * 2400L) / TotalSteps * (F_CPU / 40000) / MAXIMUM_FEEDRATE_E) << 8

    IF(ClockLimitCalculation > ClockLimit)
                
      ClockLimit := ClockLimitCalculation

    'Yes, this assumes always the x axis as the critical one regarding acceleration.
    'If we want to implement per-axis acceleration, things get tricky ...
    ClockMinimum := (MoveDuration / F_EndPointMantissa) << 8

    IF(ClockMinimum < ClockLimit)
                
      ClockMinimum := ClockLimit
      F_EndPointDDA := MoveDuration / (ClockMinimum >> 8)                

    'Lookahead can deal with 16 bits ( := 1092 mm/s), only.
    IF(F_EndPointDDA > 65535)
                
      F_EndPointDDA := 65535                

    'Acceleration ramps are based on the fast axis, not the combined speed.
    RampUpSteps := AccelerationRampLength(MulDiv(FastAxisMantissaDifference, F_EndPointDDA, Distance), FastAxisStepsPerMeter)

    IF(RampUpSteps > TotalSteps / 2)
                
      RampUpSteps := TotalSteps / 2
                

    RampDownSteps := TotalSteps - RampUpSteps

    StepTimeOffset := 0
    ClockDDA := C0

    ClockDDA := (MoveDuration / F_EndPointMantissa) << 8

    IF(ClockDDA < ClockLimit)
                
      ClockDDA := ClockLimit                

    '!TotalSteps == 0

  'Next dda starts where we finish
  memcpy(&startpoint, target, sizeof(TARGET))

PRI StartDDA(DDA *dda)

{
  Start a prepared DDA
  Parameter is a *dda pointer to entry in dda_queue to start
  This function actually begins the move described by the passed DDA entry.
  We set direction and enable outputs, and set the timer for the first step from the precalculated value.
  We also mark this DDA as running, so other parts of the firmware know that something is happening
  Called both inside and outside of interrupts.
}

  'Called from interrupt context: keep it simple!

  IF(!NullMove)
        
    'Get ready to go
    PowerSupplyTimeout := 0

    IF(Z_StepsDifference)
                
      Z_Enable()
                

    IF(EndStopCheckDDA)
                
      EndStopsOn()
                

    'Set direction outputs
    X_SetDirection(X_Direction)
    Y_SetDirection(Y_Direction)
    Z_SetDirection(Z_Direction)
    E_SetDirection(E_Direction)

    #ifdef  DC_EXTRUDER

    IF(E_StepsDifference)
                
      HeaterSet(DC_EXTRUDER, DC_EXTRUDER_PWM)
                

    #endif

    'Initialise state variable
    X_CounterMoveState := Y_CounterMoveState := Z_CounterMoveState := E_CounterMoveState := -(TotalSteps >> 1)
    memcpy(&X_StepsMoveState, &X_StepsDifference, sizeof(uint32_t) * 4)
    EndStopMoveState := 0

    StepNumberMoveState := 0

    'Ensure this dda starts
    LiveDDA := 1

    'Set timeout for first step
    SetTimer(ClockDDA >> 8)
        

  'Else just a speed change, keep LiveDDA := 0

  F_CurrentPosition := F_EndPointDDA

PRI StepDDA(DDA *dda)

{
  Make a STEP
  Parameter is a *dda the current move
  This is called from our timer interrupt every time a step needs to occur. Keep it as simple as possible!
  We first work out which axes need to step, and generate step pulses for them
  Then we re-enable global interrupts so serial data reception and other important things can occur while we do some math.
  Next, we work out how long until our next step using the selected acceleration algorithm and set the timer.
  Then we decide IF this was the last step for this move, and IF so mark this dda as dead so next timer interrupt we can start a new one.
  Finally we de-assert any asserted step pins.
}

  IF(AxisToStepDDA == "X")
        
    X_Step()
    X_StepsMoveState--
    X_TimeMoveState += X_StepIntervalDDA
    AllTimeMoveState := X_TimeMoveState
        

  IF(Y_StepsMoveState)
        
    Y_CounterMoveState -= Y_StepsDifference

    IF(Y_CounterMoveState < 0)
                
      Y_Step()
      Y_StepsMoveState--
      Y_CounterMoveState += TotalSteps                

  IF(Z_StepsMoveState)
        
    Z_CounterMoveState -= Z_StepsDifference

    IF(Z_CounterMoveState < 0)
                
      Z_Step()
      Z_StepsMoveState--
      Z_CounterMoveState += TotalSteps        

  IF(E_StepsMoveState)
        
    E_CounterMoveState -= E_StepsDifference

    IF(E_CounterMoveState < 0)
                
      E_Step()
      E_StepsMoveState--
      E_CounterMoveState += TotalSteps        

  StepNumberMoveState++

  'If there are no steps left or an endstop stop happened, we have finished.
  IF((X_StepsMoveState == 0 && Y_StepsMoveState == 0 && Z_StepsMoveState == 0 && E_StepsMoveState == 0) || (EndStopMoveState && StepTimeOffset == 0))
        
    LiveDDA := 0
    DoneDDA := 1            

    #ifdef DC_EXTRUDER

    HeaterSet(DC_EXTRUDER, 0)

    #endif

    'Z stepper is only enabled while moving
    Z_Disable()
        
  ELSE
        
    PowerSupplyTimeout := 0

    'After having finished, StartDDA() will set the timer.
    SetTimer(ClockDDA >> 8)        

  'Turn off step outputs, hopefully they've been on long enough by now to register with the drivers,
  'IF not, too bad. Or insert a (very!) small delay here, or fire up a spare timer or something.
  'we also hope that we don't step before the drivers register the low- limit maximum speed if you
  'think this is a problem.
  Unstep()

PRI ClockDDA()

{
  Do regular movement maintenance.
  This should be called pretty often, like once every 1 or 2 milliseconds.

  Currently, this is checking the endstops and doing acceleration maths. These
  dont need to be checked/recalculated on every single step, so this code
  can be moved out of the highly time critical StepDDA(). At high precision
  (slow) searches of the endstop, this function is called more often than
  StepDDA() anyways.

  In the future, arc movement calculations might go here, too. Updating
  movement direction 500 times a second is easily enough for smooth and
  accurate curves!
}

  static volatile uint8_t Busy := 0
  DDA *dda
  static DDA *LastDDA := NULL
  uint8_t EndStopTrigger := 0
  uint32_t MoveStepNumber
  uint32_t MoveClock
  uint8_t RecalculateSpeed

  dda := QueueCurrentMovement()

  'Per Bruce:  I do not know how these debounces are achieved, because the 
  'MOVE_STATE structure within dda.h do not have similar variables.
  IF(dda != LastDDA)
        
    X_DebounceMinMoveState := 0
    X_DebounceMaxMoveState := 0
    Y_DebounceMinMoveState := 0
    Y_DebounceMaxMoveState := 0
    Z_DebounceMinMoveState := 0
    Z_DebounceMaxMoveState := 0
    
    LastDDA := dda        

  IF(dda == NULL)
        
    RETURN        

  'Lengthy calculations ahead!
  'Make sure we didn't re-enter, then allow nested interrupts.
  IF(Busy)
        
    RETURN        
        
  Busy := 1
        
  Sei()

  'For maths about stepper speed profiles, see
  'http:'www.embedded.com/columns/technicalinsights/56800129?printable=true
  'and http:'www.atmel.com/images/doc8017.pdf (Atmel app note AVR446)

  ATOMIC_START

  MoveStepNumber := StepNumberMoveState

  'All other variables are read-only or unused in StepDDA(),
  'so no need for atomic operations.

  ATOMIC_END

  RecalculateSpeed := 0

  IF(MoveStepNumber < RampUpSteps)
        
    StepTimeOffset := MoveStepNumber

    RecalculateSpeed := 1
        
  ELSEIF(MoveStepNumber >= RampDownSteps)
        
    StepTimeOffset := TotalSteps - MoveStepNumber

    RecalculateSpeed := 1
    

  IF(RecalculateSpeed)
        
    IF(StepTimeOffset == 0)
                
      MoveClock := C0
                
    ELSE
                
      'Explicit formula: c0 * (sqrt(n + 1) - sqrt(n)),
      'approximation here: c0 * (1 / (2 * sqrt(n))).
      MoveClock := ((C0 >> 8) * int_InverseSquareRoot(StepTimeOffset)) >> 5
                

    'TODO: most likely this whole check is obsolete. It was left as a
    'safety margin, only. Rampup steps calculation should be accurate
    'now and give the requested target speed within a few percent.
    IF(MoveClock < ClockMinimum)
                
      'We hit max speed not always exactly.
      MoveClock := ClockMinimum                       
                

    'Write results.
    ATOMIC_START

    ClockDDA := MoveClock

    ATOMIC_END    

  Cli() ' Compensate Sei() above.
  Busy := 0

PRI UpdateCurrentPosition()

{
 Update global current_position struct
}

  DDA *dda := &MoveBuffer[MoveBufferTail]

  IF(QueueEmpty())
        
    X_CurrentPosition := X_StartPointMantissa
    Y_CurrentPosition := Y_StartPointMantissa
    Z_CurrentPosition := Z_StartPointMantissa
    E_CurrentPosition := E_StartPointMantissa
        
  ELSEIF(LiveDDA)
        
    IF(X_Direction)
                
      '(X_STEPS_PER_METER / 1000) is a bit inaccurate for low STEPS_PER_M numbers
      'should be: X_StepsMoveState * 1000000 / X_STEPS_PER_METER)
      'but x_steps can be like 1000000 already, so we'd overflow
      X_CurrentPosition := X_EndPointDDA - X_StepsMoveState * 1000 / ((X_STEPS_PER_METER + 500) / 1000)
                
    ELSE
                
      X_CurrentPosition := X_EndPointDDA + X_StepsMoveState * 1000 / ((X_STEPS_PER_METER + 500) / 1000)
                

    IF(Y_Direction)
                
      Y_CurrentPosition := Y_EndPointDDA - Y_StepsMoveState * 1000 / ((Y_STEPS_PER_METER + 500) / 1000)
                
    ELSE
                
      Y_CurrentPosition := Y_EndPointDDA + Y_StepsMoveState * 1000 / ((Y_STEPS_PER_METER + 500) / 1000)
                

    IF(Z_Direction)
                
      Z_CurrentPosition := Z_EndPointDDA - Z_StepsMoveState * 1000 / ((Z_STEPS_PER_METER + 500) / 1000)
                
    ELSE
                
      Z_CurrentPosition := Z_EndPointDDA + Z_StepsMoveState * 1000 / ((Z_STEPS_PER_METER + 500) / 1000)
                

    IF(E_EndPointDDA_relative)
                
      E_CurrentPosition := E_StepsMoveState * 1000 / ((E_STEPS_PER_METER + 500) / 1000)
                
    ELSE
                
      IF(E_Direction)
                        
        E_CurrentPosition := E_EndPointDDA - E_StepsMoveState * 1000 / ((E_STEPS_PER_METER + 500) / 1000)
                        
      ELSE
                        
        E_CurrentPosition := E_EndPointDDA + E_StepsMoveState * 1000 / ((E_STEPS_PER_METER + 500) / 1000)
                        
        'F_CurrentPosition is updated in StartDDA()        